import argparse
import collections
from redisclu.node import Node
from redisclu.cluster import Cluster
from redisclu.utils import echo, spread, divide


class Context(object):
    def __init__(self, parser=None):
        self.__parser = parser

    def abort(self, message):
        if not self.__parser:
            return

        self.__parser.error(message)


class Command(object):
    def __init__(self, args, func):
        self.arguments = args
        self.parser = None
        self.name = func.__name__
        self.doc = func.__doc__
        self.func = func

    def callback(self, args):
        ctx = Context(self.parser)
        if hasattr(self.func, '__pass_ctx__'):
            self.func(ctx, args)
        else:
            self.func(args)

    @classmethod
    def command(cls, func):
        if not hasattr(func, '__cmd_args__'):
            func.__cmd_args__ = []
        func.__cmd_args__.reverse()
        return cls(func.__cmd_args__, func)

    @classmethod
    def pass_ctx(cls, func):
        func.__pass_ctx__ = True
        return func

    @classmethod
    def argument(cls, *args, **kwargs):
        def deco(func):
            if not hasattr(func, '__cmd_args__'):
                func.__cmd_args__ = []
            func.__cmd_args__.append((args, kwargs))
            return func

        return deco

    def __call__(self):
        self.parser = argparse.ArgumentParser()
        for args, kwargs in self.arguments:
            self.parser.add_argument(*args, **kwargs)
        args = self.parser.parse_args()
        self.callback(args)


class CommandParser(object):
    def __init__(self, *args, **kwargs):
        self.parser = argparse.ArgumentParser(*args, **kwargs)
        self.subparser = self.parser.add_subparsers(title='Subcommands')

    def add_command(self, command):
        parser = self.subparser.add_parser(command.name, help=command.doc)
        command.parser = parser
        for args, kwargs in command.arguments:
            parser.add_argument(*args, **kwargs)
        parser.set_defaults(func=command.callback)

    def run(self):
        args = self.parser.parse_args()
        args.func(args)


command = Command.command
argument = Command.argument
pass_ctx = Command.pass_ctx


class MasterCandidate(object):
    def __init__(self, master):
        self.master = master
        self.unbinded_slots = []

    def __getattr__(self, attr):
        return getattr(self.master, attr)

    def bind_slots(self):
        assert self.unbinded_slots
        for chunk in self.unbinded_slots:
            self.add_slots(*range(*chunk))

    def is_enabled(self):
        return self.unbinded_slots


class ClusterCreator(object):
    def __init__(self, masters):
        master_candidates = [Node.from_uri(i) for i in masters]
        self.master_candidates = [MasterCandidate(i) for i in
                                  master_candidates]
        self.masters = []
        self.cluster = None

    def check(self):
        # check pre-requirements
        if len(self.master_candidates) < 2:
            return False
        for master in self.master_candidates:
            if not master.info().get('cluster_enabled'):
                return False
            master.execute_command('select', '0')
            if master.randomkey():
                return False
            if master.cluster_info()['cluster_known_nodes'] != 1:
                return False
        return True

    def initialize_slots(self):
        ips = collections.defaultdict(list)
        for candidate in self.master_candidates:
            ips[candidate.host].append(candidate)
        master_count = len(self.master_candidates)
        self.masters = masters = spread(ips, master_count)
        chunks = self.split_slot(Cluster.CLUSTER_HASH_SLOTS, master_count)
        for master, chunk in zip(masters, chunks):
            master.unbinded_slots.append(chunk)

        self.master_candidates = [i for i in self.master_candidates if
                                  i.is_enabled()]
        self.cluster = Cluster(self.master_candidates)

    def show_cluster_info(self):
        for instance in self.master_candidates:
            echo('M', end='', color='green')
            name_msg = ': {name} {host}:{port}'
            echo(name_msg.format(name=instance.name,
                                 host=instance.host, port=instance.port))
            slot_msg = ','.join(['-'.join([str(s[0]), str(s[1] - 1)])
                                 for s in instance.unbinded_slots])
            echo('\tslots:', slot_msg)

    def bind_slots(self):
        for master in self.masters:
            master.bind_slots()

    def join_cluster(self):
        if not self.masters:
            return

        first_master = self.masters[0]
        for master in self.masters[1:]:
            master.meet(first_master.host, first_master.port)

    def bind_config_epoch(self):
        epoch = 1
        for instance in self.masters:
            try:
                instance.set_config_epoch(epoch)
            except:
                pass
            epoch += 1

    def split_slot(self, n, m):
        chunks = divide(n, m)

        res, total = [], 0
        for c in chunks:
            res.append((total, c + total))
            total += c
        return res



from helper import CommandParser
from commands import load_commands


def main():
    parser = CommandParser(fromfile_prefix_chars='@')
    for command in load_commands():
        parser.add_command(command)

    parser.run()


if __name__ == "__main__":
    main()



import time
import itertools
import hashlib
import failover

from node import Node
from exceptions import (ClusterNotHealthy, ClusterNotConsistent)
from utils import (divide, echo)


class Cluster(object):
    CLUSTER_HASH_SLOTS = 16384

    def __init__(self, nodes, hash_slots=CLUSTER_HASH_SLOTS,
                 parent_nodes=None):
        self.nodes = nodes
        self.parent_nodes = parent_nodes if parent_nodes else nodes
        self.CLUSTER_HASH_SLOTS = hash_slots
        self.attempts = []
        self.key_migration_count = 1

    @classmethod
    def from_node(cls, node):
        nodes = [Node.from_uri(i['addr']) for i in node.nodes()
                 if i['link_status'] != 'disconnected']
        return cls(nodes)

    @property
    def masters(self):
        return [i for i in self.nodes if i.is_master()]

    def set_key_migration_count(self, val):
        self.key_migration_count = val

    def consistent(self):
        sig = set()
        for instance in self.nodes:
            if not instance.is_master():
                continue
            nodes = instance.nodes()
            slots, names = [], []
            for node in nodes:
                slots.extend(node['slots'])
                names.append(node['name'])
            info = '{}:{}'.format('|'.join(sorted(names)),
                                  ','.join(str(i) for i in sorted(slots)))
            sig.add(hashlib.md5(info).hexdigest())
        return len(sig) == 1

    def healthy(self):
        slots = list(itertools.chain(*[i.slots for i in self.nodes]))
        return len(slots) == self.CLUSTER_HASH_SLOTS and self.consistent()

    def wait(self):
        check = 0
        while not self.consistent():
            time.sleep(1)
            if check == 10:
                raise ClusterNotConsistent('Error: cluster is not consistent')
            check += 1

        if not self.healthy():
            raise ClusterNotHealthy('Error: missing slots')

    def get_node(self, node_id):
        for i in self.nodes:
            if i.name == node_id:
                return i

    def fix_open_slots(self):
        for master in self.masters:
            self.fix_node(master)

    def fix_node(self, node):
        info = node.node_info

        for slot, dst_id in info['migrating'].items():
            dst = self.get_node(dst_id)
            if not dst or slot not in dst.node_info['importing']:
                node.set_slot('STABLE', slot)
                continue

            node.migrate_slot(dst, slot, self)

        for slot, target_id in info['importing'].items():
            src = self.get_node(target_id)
            if not src or slot not in src.node_info['migrating']:
                node.set_slot('STABLE', slot)
                continue

            src.migrate_slot(node, slot, self)

    @failover.on_timeout
    def reshard(self):
        if not self.consistent():
            return

        nodes = [{
                     "node": n,
                     "count": len(n.slots),
                     "need": []
                 } for n in self.masters]

        nodes = self.slot_balance(nodes)

        for n in nodes:
            if not n["need"]:
                continue
            for src, count in n["need"]:
                self.migrate(src, n["node"], count)

    @failover.on_timeout
    def remove_node(self, node):
        if node.is_master():
            self.migrate_node(node)

        self.nodes = [n for n in self.nodes if n.name != node.name]
        masters = self.masters
        masters.sort(key=lambda x: len(x.slaves(x.name)))

        for n in self.nodes:
            if n.is_slave(node.name):
                n.replicate(masters[0].name)
            n.forget(node.name)

        assert not node.slots
        node.reset()

    def add_node(self, master):
        """
        Add node to cluster.
        """
        new = Node.from_uri(master)
        cluster_member = self.nodes[0]
        new.meet(cluster_member.host, cluster_member.port)
        self.nodes.append(new)
        self.wait()

    def fill_slots(self):
        masters = self.masters
        slots = itertools.chain(*[n.slots for n in masters])
        missing = list(set(range(self.CLUSTER_HASH_SLOTS)).difference(slots))

        div = divide(len(missing), len(masters))
        masters.sort(key=lambda x: len(x.slots))

        i = 0
        for count, node in zip(div, masters):
            node.add_slots(*missing[i:count + i])
            i += count

    def migrate_node(self, src_node):
        nodes = [n for n in self.masters if n.name != src_node.name]
        slot_count = len(src_node.slots)
        if slot_count <= 0:
            return
        slots = divide(slot_count, len(nodes))

        nodes.sort(key=lambda x: len(x.slots))

        for node, count in zip(nodes, slots):
            src, dst = (src_node, node)
            self.migrate(src, dst, count)

    def migrate(self, src, dst, count):
        if count <= 0:
            return

        slots = src.slots
        slots_count = len(slots)
        if count > slots_count:
            count = slots_count

        keys = [(s, src.count_keys_in_slot(s)) for s in slots]
        keys.sort(key=lambda x: x[1])

        for slot, _ in keys[:count]:
            src.migrate_slot(dst, slot, self)

    def update_slot_mapping(self, slot, dst_name):
        for node in self.parent_nodes:
            node.set_slot('NODE', slot, dst_name)

    def print_attempts(self):
        for node in self.nodes:
            self.attempts.extend(node.attempts)
        echo('Length of attempts: {}'.format(len(self.attempts)))
        for exc, group in itertools.groupby(self.attempts,
                                            lambda a: type(a).__name__):
            length = len(list(group))
            echo(
                'Exception: {}, Count: {}'.format(exc, length),
                color='yellow')

    def slot_balance(self, seq):
        amt = self.CLUSTER_HASH_SLOTS
        seq.sort(key=lambda x: x['count'], reverse=True)
        chunks = divide(amt, len(seq))
        pairs = list(zip(seq, chunks))

        i, j = 0, len(pairs) - 1
        while i < j:
            m, count = pairs[i]
            more = m['count'] - count
            if more <= 0:
                i += 1
                continue

            n, count = pairs[j]
            need = count - n['count']
            if need <= 0:
                j -= 1
                continue

            if need < more:
                n['need'].append((m['node'], need))
                n['count'] += need
                m['count'] -= need
                j -= 1
            elif need > more:
                n['need'].append((m['node'], more))
                n['count'] += more
                m['count'] -= more
                i += 1
            else:
                n['need'].append((m['node'], need))
                n['count'] += need
                m['count'] -= more
                j -= 1
                i += 1

        return seq



from redis import connection
from redis.exceptions import (ResponseError)
from exceptions import (AskError, MovedError)

import logging

logging.basicConfig()

CUSTOM_EXCEPTION_CLASSES = {
    'ASK': AskError,
    'MOVED': MovedError,
}


def parse_error(self, response):
    "Parse an error response"
    error_code = response.split(' ')[0]
    if error_code in self.EXCEPTION_CLASSES:
        response = response[len(error_code) + 1:]
        exception_class = self.EXCEPTION_CLASSES[error_code]
        if isinstance(exception_class, dict):
            for reason, inner_exception_class in exception_class.items():
                if reason in response:
                    return inner_exception_class(response)
            return ResponseError(response)
        return exception_class(response)
    return ResponseError(response)


connection.BaseParser.EXCEPTION_CLASSES.update(CUSTOM_EXCEPTION_CLASSES)
connection.BaseParser.parse_error = parse_error



from redis.exceptions import (
    ResponseError
)


class RedisCluException(Exception):
    pass


class AskError(ResponseError):
    """
    partially keys is slot migrated to another node

    src node: MIGRATING to dst node
        get > ASK error
        ask dst node > ASKING command
    dst node: IMPORTING from src node
        asking command only affects next command
        any op will be allowed after asking command
    """

    def __init__(self, resp):
        """should only redirect to master node"""
        self.args = (resp,)
        self.message = resp
        slot_id, new_node = resp.split(' ')
        host, port = new_node.rsplit(':', 1)
        self.slot_id = int(slot_id)
        self.node_addr = self.host, self.port = host, int(port)


class MovedError(AskError):
    """
    all keys in slot migrated to another node
    """
    pass


class ClusterNotHealthy(RedisCluException):
    pass


class ClusterNotConsistent(RedisCluException):
    pass


from __future__ import print_function

import itertools
import os
import sys

COLOR_MAP = {
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "purple": 35
}


def echo(*values, **kwargs):
    end = kwargs.get("end", '\n')
    color = kwargs.get("color", None)
    bold = 0 if kwargs.get("bold", False) is False else 1
    disable = kwargs.get("diable", False)

    if disable:
        return

    msg = ' '.join(str(v) for v in values) + end

    if not color or os.getenv("ANSI_COLORS_DISABLED") is not None:
        sys.stdout.write(msg)
    else:
        color_prefix = "\033[{};{}m".format(bold, COLOR_MAP[color])
        color_suffix = "\033[0m"
        sys.stdout.write(color_prefix + msg + color_suffix)
    sys.stdout.flush()


def divide(n, m):
    """Divide integer n to m chunks
    """
    avg = int(n / m)
    remain = n - m * avg
    data = list(itertools.repeat(avg, m))
    for i in range(len(data)):
        if not remain:
            break
        data[i] += 1
        remain -= 1
    return data


def spread(nodes, n):
    """Distrubute master instances in different nodes

    {
        "192.168.0.1": [node1, node2],
        "192.168.0.2": [node3, node4],
        "192.168.0.3": [node5, node6]
    } => [node1, node3, node5]
    """
    target = []

    while len(target) < n and nodes:
        for ip, node_group in list(nodes.items()):
            if not node_group:
                nodes.pop(ip)
                continue
            target.append(node_group.pop(0))
            if len(target) >= n:
                break
    return target


import time
from redis.exceptions import TimeoutError

from utils import echo


def on_timeout(f):
    def wrapper(*args, **kwargs):
        for _ in range(100):
            try:
                return f(*args, **kwargs)
            except TimeoutError as e:
                args[0].attempts.append(e)
                echo('Timeout error reading from socket. '
                     'Trying again in 10 seconds.',
                     color='red')
                time.sleep(10)
        raise

    return wrapper



import sys
import time
import redis
import itertools
import concurrent.futures

from random import shuffle

from redisclu.cli import helper as cli_helper
from redisclu.cluster import Cluster
from redisclu.node import Node, MockNode
from redisclu.utils import echo


@cli_helper.command
@cli_helper.argument('masters', nargs='+')
def create(args):
    creator = cli_helper.ClusterCreator(args.masters)
    if not creator.check():
        echo('Pre-requirements to create cluster.\n', color='red')
        echo('\t1. At least 2 redis instances must be provided.')
        echo('\t2. Should be set <cluster_enabled> on redis server conf.')
        echo('\t3. Should be removed all keys in db 0.')
        echo('\t4. Any redis instance should not be member of other cluster.')
        exit()
    creator.initialize_slots()
    creator.show_cluster_info()
    creator.bind_slots()
    creator.bind_config_epoch()
    creator.join_cluster()
    echo('Waiting for the cluster to join ', end='')
    sys.stdout.flush()
    time.sleep(1)
    while not creator.cluster.consistent():
        echo('.', end='')
        sys.stdout.flush()
        time.sleep(1)


@cli_helper.command
@cli_helper.argument('cluster')
def status(args):
    cluster = Cluster.from_node(Node.from_uri(args.cluster))
    dis = []
    for n in cluster.masters:
        slaves = ','.join([s['addr'] for s in n.slaves(n.name)])
        msg = '{} {}:{} {} {}'.format(n.name, n.host, n.port, len(n.slots),
                                      slaves)
        dis.append(msg)
    echo('\n'.join(dis))
    echo('Masters:', len(cluster.masters))
    echo('Slaves:', len(cluster.nodes) - len(cluster.masters))
    covered_slots = sum(len(n.slots) for n in cluster.masters)
    echo('Covered Slots:', covered_slots)
    if covered_slots == cluster.CLUSTER_HASH_SLOTS:
        echo('Cluster is healthy!')
    else:
        echo('!!!Cluster is not healthy!!!')
        echo('Either there is no cluster or exists cluster is not healthy.')
        echo('If there is no cluster then run "redis-clu create <master_list>" or')
        echo('"redis-clu fix {}" would be great if there is exists cluster!'.format(args.cluster))

    echo('\n')

    for master in cluster.masters:
        echo(master)
        echo('===========================')
        echo(master.execute_command('info', 'keyspace'))
        echo('\n')


@cli_helper.command
@cli_helper.argument('cluster')
def fix(args):
    cluster = Cluster.from_node(Node.from_uri(args.cluster))
    cluster.fix_open_slots()
    cluster.fill_slots()
    cluster.wait()
    cluster.print_attempts()


@cli_helper.command
@cli_helper.argument('cluster')
@cli_helper.argument('master')
@cli_helper.argument('--keyMigrationCount', default=1)
@cli_helper.pass_ctx
def add(ctx, args):
    """
    add master node to cluster
    """
    cluster = Cluster.from_node(Node.from_uri(args.cluster))
    if not cluster.healthy():
        ctx.abort(
            'Cluster not healthy. Run "redis-clu fix {}" first'.format(
                args.cluster))
    cluster.set_key_migration_count(int(args.keyMigrationCount))
    cluster.add_node(args.master)
    cluster.reshard()
    cluster.wait()
    cluster.print_attempts()


@cli_helper.command
@cli_helper.argument('cluster')
@cli_helper.argument('masters', nargs='+')
@cli_helper.argument('--keyMigrationCount', default=1)
@cli_helper.pass_ctx
def add_multi(ctx, args):
    cluster = Cluster.from_node(Node.from_uri(args.cluster))
    if len(args.masters) > len(cluster.nodes):
        ctx.abort('Length of new node list should be less than length of '
                  'cluster master nodes.')
    if not cluster.healthy():
        ctx.abort(
            'Cluster not healthy. Run "redis-clu fix {}" first'.format(
                args.cluster))
    masters = filter(lambda n: n.is_master(), cluster.nodes)
    residual_count = len(masters) % len(args.masters)
    if residual_count:
        for i in range(len(args.masters) - residual_count):
            masters.append(MockNode())

    for master in args.masters:
        cluster.add_node(master)

    shard_ratio = len(masters) / len(args.masters)

    shuffle(masters)

    sub_clusters = []
    while masters:
        sub_nodes = masters[:shard_ratio]
        new_master = args.masters.pop()
        nodes = filter(lambda n: not isinstance(n, MockNode), sub_nodes)
        nodes.append(Node.from_uri(new_master))
        hash_slots = len(list(itertools.chain(
            *[n.slots for n in nodes])))
        sub_cluster = Cluster(nodes, hash_slots=hash_slots,
                              parent_nodes=cluster.nodes)
        sub_cluster.set_key_migration_count(int(args.keyMigrationCount))
        sub_clusters.append(sub_cluster)
        for sn in sub_nodes:
            masters.pop(masters.index(sn))

    future_to_args = dict()
    executor = concurrent.futures.ThreadPoolExecutor(
        max_workers=len(sub_clusters))

    for sub_cluster in sub_clusters:
        args = ()
        future = executor.submit(sub_cluster.reshard)
        future_to_args.setdefault(future, args)

    concurrent.futures.wait(future_to_args)
    executor.shutdown(wait=False)
    time.sleep(1)
    cluster.wait()

    for sub_cluster in sub_clusters:
        sub_cluster.print_attempts()


@cli_helper.command
@cli_helper.argument('cluster')
@cli_helper.argument('--keyMigrationCount', default=1)
@cli_helper.pass_ctx
def reshard(ctx, args):
    cluster = Cluster.from_node(Node.from_uri(args.cluster))
    if not cluster.healthy():
        ctx.abort(
            'Cluster not healthy. Run "redis-clu fix {}" first'.format(
                args.cluster))
    cluster.set_key_migration_count(int(args.keyMigrationCount))
    cluster.reshard()
    cluster.wait()
    cluster.print_attempts()


@cli_helper.command
@cli_helper.argument('cluster')
@cli_helper.argument('node')
@cli_helper.argument('--keyMigrationCount', default=1)
@cli_helper.pass_ctx
def remove(ctx, args):
    '''
    remove node from cluster
    '''
    cluster = Cluster.from_node(Node.from_uri(args.cluster))
    if not cluster.healthy():
        ctx.abort(
            'Cluster not healthy. Run "redis-clu fix {}" first'.format(
                args.cluster))
    cluster.set_key_migration_count(int(args.keyMigrationCount))
    cluster.remove_node(Node.from_uri(args.node))
    cluster.wait()
    cluster.print_attempts()


@cli_helper.command
@cli_helper.argument('master')
@cli_helper.argument('slave')
@cli_helper.pass_ctx
def replicate(ctx, args):
    master = Node.from_uri(args.master)
    if not master.is_master():
        ctx.abort('Node {} is not a master.'.format(args.master))
    cluster = Cluster.from_node(master)
    cluster.add_node(args.slave)
    slave = Node.from_uri(args.slave)
    try:
        slave.replicate(master.name)
    except redis.ResponseError as e:
        ctx.abort(str(e))
    cluster.wait()


@cli_helper.command
@cli_helper.argument('cluster')
@cli_helper.argument('--hard', default=0)
def reset(args):
    cluster = Cluster.from_node(Node.from_uri(args.cluster))
    future_to_args = dict()
    executor = concurrent.futures.ThreadPoolExecutor(
        max_workers=len(cluster.masters))

    for master in cluster.masters:
        f_args = ()
        future = executor.submit(master.flushall)
        future_to_args.setdefault(future, f_args)

    concurrent.futures.wait(future_to_args)
    executor.shutdown(wait=False)

    if int(args.hard) == 1:
        for node in cluster.nodes:
            node.reset(hard=True)


def load_commands():
    import commands

    attrs = set(dir(commands))
    return filter(lambda f: isinstance(f, cli_helper.Command),
                  map(lambda attr: getattr(commands, attr), attrs))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

version = '0.0.4'

setup(
    name='redis-clu',
    version=version,
    description='Redis Cluster Management Tool',
    long_description=README,
    url='https://github.com/baranbartu/redis-clu',
    download_url='https://github.com/baranbartu/redis-clu/tarball/%s' % version,
    author='Baran Bartu Demirci',
    author_email='bbartu.demirci@gmail.com',
    license='MIT',
    keywords='redis cluster management replication master slave',
    packages=['redisclu'],
    entry_points={
        'console_scripts': ['redis-clu = redisclu.cli:main']
    },
    install_requires=[
        'hiredis',
        'redis',
        'futures==3.0.3',
    ]
)



import socket
import redis
import urlparse

from exceptions import (AskError, MovedError)
from utils import echo


class Node(object):
    ignored_exceptions = (AskError, MovedError)

    def __init__(self, host='localhost', port=6379, socket_timeout=4):
        self.host = socket.gethostbyname(host)
        self.port = port
        self.redis = redis.Redis(host, port, socket_timeout=socket_timeout)
        self.redis.ping()
        self.pipeline = self.redis.pipeline(transaction=False)
        self.attempts = []

    @classmethod
    def from_uri(cls, uri):
        if not uri.startswith('redis://'):
            uri = 'redis://{}'.format(uri)
        d = urlparse.urlparse(uri)
        return cls(d.hostname, d.port)

    def __repr__(self):
        return 'Node<{}:{}>'.format(self.host, self.port)

    def __getattr__(self, attr):
        return getattr(self.redis, attr)

    def execute_command(self, *args, **kwargs):
        return self.redis.execute_command(*args, **kwargs)

    def is_slave(self, master_id=None):
        info = self.node_info

        r = 'slave' in info['flags']
        if master_id is not None:
            r = r and info['replicate'] == master_id
        return r

    def is_master(self):
        return 'master' in self.node_info['flags']

    @property
    def node_info(self):
        nodes = self.nodes()
        return nodes[0]

    @property
    def slots(self):
        return self.node_info['slots']

    @property
    def name(self):
        return self.node_info['name']

    def migrate_keys(self, host, port, keys):
        for key in keys:
            self.pipeline.execute_command('MIGRATE', host, port, key, 0, 15000)
        return self.pipeline.execute(raise_on_error=False)

    def migrate_slot(self, dst, slot, cluster):
        dst.set_slot('IMPORTING', slot, self.name)
        self.set_slot('MIGRATING', slot, dst.name)

        total_keys = 0
        for keys in self._scan_keys(slot, cluster.key_migration_count):
            results = self.migrate_keys(dst.host, dst.port, keys)
            self.attempts.extend(filter(lambda r: any(
                isinstance(r, e) for e in self.ignored_exceptions), results))
            total_keys += len(keys)

        echo('{} key(s) migrated from {} to {} in slot {}'.format(
            total_keys, self, dst, slot))

        cluster.update_slot_mapping(slot, dst.name)

    def reset(self, hard=False):
        args = []
        if hard:
            args = ['HARD']
        return self.execute_command('CLUSTER RESET', *args)

    def set_slot(self, action, slot, node_id=None):
        remain = [node_id] if node_id else []
        return self.execute_command('CLUSTER SETSLOT', slot, action, *remain)

    def get_keys_in_slot(self, slot, count):
        return self.execute_command('CLUSTER GETKEYSINSLOT', slot, count)

    def count_keys_in_slot(self, slot):
        return self.execute_command('CLUSTER COUNTKEYSINSLOT', slot)

    def slaves(self, node_id):
        data = self.execute_command('CLUSTER SLAVES', node_id)
        return self._parse_node('\n'.join(data))

    def add_slots(self, *slot):
        if not slot:
            return
        self.execute_command('CLUSTER ADDSLOTS', *slot)

    def forget(self, node_id):
        return self.execute_command('CLUSTER FORGET', node_id)

    def set_config_epoch(self, config_epoch):
        return self.execute_command('CLUSTER SET-CONFIG-EPOCH', config_epoch)

    def meet(self, ip, port):
        return self.execute_command('CLUSTER MEET', ip, port)

    def replicate(self, node_id):
        return self.execute_command('CLUSTER REPLICATE', node_id)

    def nodes(self):
        info = self.execute_command('CLUSTER NODES').strip()
        return self._parse_node(info)

    def cluster_info(self):
        data = {}
        info = self.execute_command('CLUSTER INFO').strip()
        for item in info.split('\r\n'):
            k, v = item.split(':')
            if k != 'cluster_state':
                v = int(v)
            data[k] = v
        return data

    def _parse_node(self, nodes):
        data = []
        for item in nodes.split('\n'):
            if not item:
                continue
            confs = item.split()
            node_info = {
                'name': confs[0],
                'addr': confs[1],
                'flags': confs[2].split(','),
                'replicate': confs[3],  # master_id
                'ping_sent': int(confs[4]),
                'ping_recv': int(confs[5]),
                'link_status': confs[7],
                'migrating': {},
                'importing': {},
                'slots': []
            }
            for slot in confs[8:]:
                if slot[0] == '[':
                    if '->-' in slot:
                        s, dst = slot[1:-1].split('->-')
                        node_info['migrating'][s] = dst
                    elif '-<-' in slot:
                        s, src = slot[1:-1].split('-<-')
                        node_info['importing'][s] = src
                elif '-' in slot:
                    start, end = slot.split('-')
                    node_info['slots'].extend(range(int(start), int(end) + 1))
                else:
                    node_info['slots'].append(int(slot))

            if 'myself' in node_info['flags']:
                data.insert(0, node_info)
            else:
                data.append(node_info)
        return data

    def _scan_keys(self, slot, count):
        while True:
            keys = self.get_keys_in_slot(slot, count)
            if not keys:
                break
            yield keys


class MockNode(object): pass
